# Copyright 2019 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from unittest import mock

from oslotest import base

from sushy_tools.emulator.resources.vmedia.staticdriver import StaticDriver


@mock.patch('sushy_tools.emulator.memoize.PersistentDict', new=dict)
class StaticDriverTestCase(base.BaseTestCase):

    UUID = 'ZZZ-YYY-XXX'

    CONFIG = {
        'SUSHY_EMULATOR_VMEDIA_DEVICES': {
            "Cd": {
                "Name": "Virtual CD",
                "MediaTypes": [
                    "CD",
                    "DVD"
                ]
            },
            "Floppy": {
                "Name": "Virtual Removable Media",
                "MediaTypes": [
                    "Floppy",
                    "USBStick"
                ]
            }
        }
    }

    def test_devices(self):
        test_driver = StaticDriver.initialize(
            self.CONFIG, mock.MagicMock())()

        devices = test_driver.devices
        self.assertEqual(['Cd', 'Floppy'], sorted(devices))

    def test_get_device_name(self):
        test_driver = StaticDriver.initialize(
            self.CONFIG, mock.MagicMock())()

        device_name = test_driver.get_device_name(
            self.UUID, 'Cd')
        self.assertEqual('Virtual CD', device_name)

    def test_get_device_media_types(self):
        test_driver = StaticDriver.initialize(
            self.CONFIG, mock.MagicMock())()

        media_types = test_driver.get_device_media_types(
            self.UUID, 'Cd')
        self.assertEqual(['CD', 'DVD'], media_types)

    def test_get_device_image_info(self):
        test_driver = StaticDriver.initialize(
            self.CONFIG, mock.MagicMock())()

        dev_info = test_driver.get_device_image_info(
            self.UUID, 'Cd')
        expected = ('', '', False, False)
        self.assertEqual(expected, dev_info)

    @mock.patch('sushy_tools.emulator.resources.vmedia'
                '.staticdriver.StaticDriver._get_device',
                autospec=True)
    @mock.patch('sushy_tools.emulator.resources.vmedia'
                '.staticdriver.open')
    @mock.patch('sushy_tools.emulator.resources.vmedia'
                '.staticdriver.os.rename', autospec=True)
    @mock.patch('sushy_tools.emulator.resources.vmedia'
                '.staticdriver.tempfile', autospec=True)
    @mock.patch('sushy_tools.emulator.resources.vmedia'
                '.staticdriver.requests', autospec=True)
    def test_insert_image(self, mock_requests, mock_tempfile, mock_rename,
                          mock_open, mock_get_device):
        test_driver = StaticDriver.initialize(
            self.CONFIG, mock.MagicMock())()

        device_info = {}
        mock_get_device.return_value = device_info

        mock_tempfile.mkdtemp.return_value = '/alphabet/soup'
        mock_tempfile.gettempdir.return_value = '/tmp'
        mock_tmp_file = (mock_tempfile.NamedTemporaryFile
                         .return_value.__enter__.return_value)
        mock_tmp_file.name = 'alphabet.soup'
        mock_rsp = mock_requests.get.return_value.__enter__.return_value
        mock_rsp.headers = {
            'content-disposition': 'attachment; filename="fish.iso"'
        }

        local_file = test_driver.insert_image(
            self.UUID, 'Cd', 'http://fish.it/red.iso', inserted=True,
            write_protected=False)

        self.assertEqual('/alphabet/soup/fish.iso', local_file)
        mock_requests.get.assert_called_once_with(
            'http://fish.it/red.iso', stream=True)
        mock_open.assert_called_once_with(mock.ANY, 'wb')
        mock_rename.assert_called_once_with(
            'alphabet.soup', '/alphabet/soup/fish.iso')

        self.assertEqual('fish.iso', device_info['Image'])
        self.assertTrue(device_info['Inserted'])
        self.assertFalse(device_info['WriteProtected'])

    @mock.patch('sushy_tools.emulator.resources.vmedia'
                '.staticdriver.StaticDriver._get_device',
                autospec=True)
    @mock.patch('sushy_tools.emulator.resources.vmedia'
                '.staticdriver.open')
    @mock.patch('sushy_tools.emulator.resources.vmedia'
                '.staticdriver.os.rename', autospec=True)
    @mock.patch('sushy_tools.emulator.resources.vmedia'
                '.staticdriver.tempfile', autospec=True)
    @mock.patch('sushy_tools.emulator.resources.vmedia'
                '.staticdriver.requests', autospec=True)
    def test_insert_image_no_local_file(self, mock_requests, mock_tempfile,
                                        mock_rename, mock_open,
                                        mock_get_device):
        test_driver = StaticDriver.initialize(
            self.CONFIG, mock.MagicMock())()

        device_info = {}
        mock_get_device.return_value = device_info

        mock_tempfile.mkdtemp.return_value = '/alphabet/soup'
        mock_tempfile.gettempdir.return_value = '/tmp'
        mock_tmp_file = (mock_tempfile.NamedTemporaryFile
                         .return_value.__enter__.return_value)
        mock_tmp_file.name = 'alphabet.soup'
        mock_rsp = mock_requests.get.return_value.__enter__.return_value
        mock_rsp.headers = {}

        local_file = test_driver.insert_image(
            self.UUID, 'Cd', 'http://fish.it/red.iso', inserted=True,
            write_protected=False)

        self.assertEqual('/alphabet/soup/red.iso', local_file)
        mock_requests.get.assert_called_once_with(
            'http://fish.it/red.iso', stream=True)
        mock_open.assert_called_once_with(mock.ANY, 'wb')
        mock_rename.assert_called_once_with(
            'alphabet.soup', '/alphabet/soup/red.iso')

        self.assertEqual('red.iso', device_info['Image'])
        self.assertTrue(device_info['Inserted'])
        self.assertFalse(device_info['WriteProtected'])

    @mock.patch('sushy_tools.emulator.resources.vmedia'
                '.staticdriver.StaticDriver._get_device',
                autospec=True)
    @mock.patch('sushy_tools.emulator.resources.vmedia'
                '.staticdriver.open')
    @mock.patch('sushy_tools.emulator.resources.vmedia'
                '.staticdriver.os.rename', autospec=True)
    @mock.patch('sushy_tools.emulator.resources.vmedia'
                '.staticdriver.tempfile', autospec=True)
    @mock.patch('sushy_tools.emulator.resources.vmedia'
                '.staticdriver.requests', autospec=True)
    def test_insert_image_full_url_v6(self, mock_requests, mock_tempfile,
                                      mock_rename, mock_open,
                                      mock_get_device):
        test_driver = StaticDriver.initialize(
            self.CONFIG, mock.MagicMock())()

        device_info = {}
        mock_get_device.return_value = device_info

        mock_tempfile.mkdtemp.return_value = '/alphabet/soup'
        mock_tempfile.gettempdir.return_value = '/tmp'
        mock_tmp_file = (mock_tempfile.NamedTemporaryFile
                         .return_value.__enter__.return_value)
        mock_tmp_file.name = 'alphabet.soup'
        mock_rsp = mock_requests.get.return_value.__enter__.return_value
        mock_rsp.headers = {}

        full_url = 'http://[::2]:80/redfish/boot-abc?filename=tmp.iso'
        local_file = test_driver.insert_image(
            self.UUID, 'Cd', full_url,
            inserted=True, write_protected=False)

        self.assertEqual('/alphabet/soup/boot-abc', local_file)
        mock_requests.get.assert_called_once_with(full_url, stream=True)
        mock_open.assert_called_once_with(mock.ANY, 'wb')
        mock_rename.assert_called_once_with(
            'alphabet.soup', '/alphabet/soup/boot-abc')

        self.assertEqual('boot-abc', device_info['Image'])
        self.assertTrue(device_info['Inserted'])
        self.assertFalse(device_info['WriteProtected'])

    @mock.patch('sushy_tools.emulator.resources.vmedia'
                '.staticdriver.StaticDriver._get_device',
                autospec=True)
    @mock.patch('sushy_tools.emulator.resources.vmedia'
                '.staticdriver.os.unlink', autospec=True)
    def test_eject_image(self, mock_unlink, mock_get_device):
        test_driver = StaticDriver.initialize(
            self.CONFIG, mock.MagicMock())()

        device_info = {
            '_local_file': '/tmp/fish.iso'
        }
        mock_get_device.return_value = device_info

        test_driver.eject_image(self.UUID, 'Cd')

        self.assertEqual('', device_info['Image'])
        self.assertEqual('', device_info['ImageName'])
        self.assertFalse(device_info['Inserted'])
        self.assertFalse(device_info['WriteProtected'])

        mock_unlink.assert_called_once_with('/tmp/fish.iso')



# Copyright 2019 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import abc

from sushy_tools.emulator import base


class AbstractManagersDriver(base.DriverBase, metaclass=abc.ABCMeta):
    """Base class backing Redfish Managers"""

    @classmethod
    def initialize(cls, config, logger, *args, **kwargs):
        cls._config = config
        cls._logger = logger
        return cls

    @abc.abstractproperty
    def driver(self):
        """Return human-friendly driver information

        :returns: driver information as `str`
        """

    @abc.abstractproperty
    def managers(self):
        """Return available Redfish managers

        :returns: list of UUIDs representing the managers
        """

    @abc.abstractmethod
    def uuid(self, identity):
        """Get Redfish manager UUID

        The universal unique identifier (UUID) for this system. Can be used
        in place of manager name if there are duplicates.

        If driver backend does not support non-unique manager identity,
        this method may just return the `identity`.

        :returns: Redfish manager UUID
        """

    @abc.abstractmethod
    def name(self, identity):
        """Get Redfish manager name by UUID

        The universal unique identifier (UUID) for this Redfish manager.
        Can be used in place of manager name if there are duplicates.

        If driver backend does not support manager names, this method may
        just return the `identity`.

        :returns: Redfish manager name
        """



# Copyright 2018 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

try:
    from collections import abc as collections

except ImportError:
    import collections

import contextlib
from functools import wraps
import os
import pickle
import sqlite3
import tempfile

# Python 3.8
MutableMapping = getattr(collections, 'abc', collections).MutableMapping


def memoize(permanent_cache=None):
    """Cache the return value of the decorated method.

    :param permanent_cache: a `dict` like object to use as a cache.
        If not given, the `._cache` attribute would be added to
        the object of the decorated method pointing to a newly
        created `dict`.
    :return: decorated function
    """

    def decorator(method):

        @wraps(method)
        def wrapped(self, *args, **kwargs):
            if permanent_cache is None:
                try:
                    cache = self._cache

                except AttributeError:
                    cache = self._cache = {}

            else:
                cache = permanent_cache

            method_cache = cache.setdefault(method, {})

            key = frozenset(args), frozenset(kwargs)

            try:
                return method_cache[key]

            except KeyError:
                rv = method(self, *args, **kwargs)
                method_cache[key] = rv
                return rv

        return wrapped

    return decorator


class PersistentDict(MutableMapping):
    DBPATH = os.path.join(tempfile.gettempdir(), 'sushy-emulator')

    _dbpath = None

    def make_permanent(self, dbpath, dbfile):
        dbpath = dbpath or self.DBPATH
        if not os.path.exists(dbpath):
            os.makedirs(dbpath)

        self._dbpath = os.path.join(dbpath, dbfile) + '.sqlite'

        with self.connection() as cursor:
            cursor.execute(
                'create table if not exists cache '
                '(key blob primary key not null, value blob not null)'
            )

        self.update(dict(self))

    @staticmethod
    def encode(obj):
        return pickle.dumps(obj)

    @staticmethod
    def decode(blob):
        return pickle.loads(blob)

    @contextlib.contextmanager
    def connection(self):
        if not self._dbpath:
            raise TypeError('Dict is not yet persistent')

        connection = sqlite3.connect(self._dbpath)

        try:
            yield connection.cursor()

            connection.commit()

        finally:
            connection.close()

    def __getitem__(self, key):
        key = self.encode(key)

        with self.connection() as cursor:
            cursor.execute(
                'select value from cache where key=?',
                (key,)
            )
            value = cursor.fetchone()

        if value is None:
            raise KeyError(key)

        return self.decode(value[0])

    def __setitem__(self, key, value):
        key = self.encode(key)
        value = self.encode(value)

        with self.connection() as cursor:
            cursor.execute(
                'insert or replace into cache values (?, ?)',
                (key, value)
            )

    def __delitem__(self, key):
        key = self.encode(key)

        with self.connection() as cursor:
            cursor.execute(
                'select count(*) from cache where key=?',
                (key,)
            )

            if cursor.fetchone()[0] == 0:
                raise KeyError(key)

            cursor.execute(
                'delete from cache where key=?',
                (key,)
            )

    def __iter__(self):
        with self.connection() as cursor:
            cursor.execute(
                'select key from cache'
            )
            records = cursor.fetchall()

        for r in records:
            yield self.decode(r[0])

    def __len__(self):
        with self.connection() as cursor:
            cursor.execute(
                'select count(*) from cache'
            )
            count = cursor.fetchone()[0]
            return count



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys
from unittest import mock
import uuid
import xml.etree.ElementTree as ET

import libvirt
from oslotest import base

from sushy_tools.emulator.resources.systems.libvirtdriver import LibvirtDriver
from sushy_tools import error


class LibvirtDriverTestCase(base.BaseTestCase):

    name = 'QEmu-fedora-i686'
    uuid = 'c7a5fdbd-cdaf-9455-926a-d65c16db1809'

    def setUp(self):
        test_driver_class = LibvirtDriver.initialize(
            {}, mock.MagicMock())
        self.test_driver = test_driver_class()
        super(LibvirtDriverTestCase, self).setUp()

    @mock.patch('libvirt.open', autospec=True)
    def test__get_domain_by_name(self, libvirt_mock):
        conn_mock = libvirt_mock.return_value
        lookupByUUID_mock = conn_mock.lookupByUUID
        domain_mock = lookupByUUID_mock.return_value
        domain_mock.UUIDString.return_value = self.uuid
        self.assertRaises(
            error.AliasAccessError, self.test_driver._get_domain, self.name)

    @mock.patch('libvirt.open', autospec=True)
    def test__get_domain_by_uuid(self, libvirt_mock):
        domain_id = uuid.UUID(self.uuid)

        conn_mock = libvirt_mock.return_value
        lookupByUUID_mock = conn_mock.lookupByUUID
        self.test_driver._get_domain(str(domain_id))
        lookupByUUID_mock.assert_called_once_with(domain_id.bytes)

    @mock.patch('libvirt.openReadOnly', autospec=True)
    def test_uuid(self, libvirt_mock):
        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByName.return_value
        domain_mock.UUIDString.return_value = self.uuid
        self.assertRaises(error.AliasAccessError,
                          self.test_driver.uuid, 'name')

    @mock.patch('libvirt.openReadOnly', autospec=True)
    def test_systems(self, libvirt_mock):
        conn_mock = libvirt_mock.return_value
        domain = mock.MagicMock()
        domain.UUIDString.return_value = self.uuid
        conn_mock.listAllDomains.return_value = [domain]
        uuidstring_mock = conn_mock.lookupByName.return_value.UUIDString
        uuidstring_mock.return_value = self.uuid
        systems = self.test_driver.systems
        self.assertEqual([self.uuid], systems)

    @mock.patch('libvirt.openReadOnly', autospec=True)
    def test_get_power_state_on(self, libvirt_mock):
        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.UUIDString.return_value = self.uuid

        domain_mock.isActive.return_value = True
        domain_mock.maxMemory.return_value = 1024 * 1024
        domain_mock.maxVcpus.return_value = 2

        power_state = self.test_driver.get_power_state(self.uuid)

        self.assertEqual('On', power_state)

    @mock.patch('libvirt.openReadOnly', autospec=True)
    def test_get_power_state_off(self, libvirt_mock):
        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.isActive.return_value = False

        power_state = self.test_driver.get_power_state(self.uuid)

        self.assertEqual('Off', power_state)

    @mock.patch('libvirt.open', autospec=True)
    def test_set_power_state_on(self, libvirt_mock):
        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.isActive.return_value = False

        self.test_driver.set_power_state(self.uuid, 'On')

        domain_mock.create.assert_called_once_with()

    @mock.patch('libvirt.open', autospec=True)
    def test_set_power_state_forceon(self, libvirt_mock):
        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.isActive.return_value = False

        self.test_driver.set_power_state(self.uuid, 'ForceOn')

        domain_mock.create.assert_called_once_with()

    @mock.patch('libvirt.open', autospec=True)
    def test_set_power_state_forceoff(self, libvirt_mock):
        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.isActive.return_value = True

        self.test_driver.set_power_state(self.uuid, 'ForceOff')

        domain_mock.destroy.assert_called_once_with()

    @mock.patch('libvirt.open', autospec=True)
    def test_set_power_state_gracefulshutdown(self, libvirt_mock):
        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.isActive.return_value = True

        self.test_driver.set_power_state(self.uuid, 'GracefulShutdown')

        domain_mock.shutdown.assert_called_once_with()

    @mock.patch('libvirt.open', autospec=True)
    def test_set_power_state_gracefulrestart(self, libvirt_mock):
        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.isActive.return_value = True

        self.test_driver.set_power_state(self.uuid, 'GracefulRestart')

        domain_mock.reboot.assert_called_once_with()

    @mock.patch('libvirt.open', autospec=True)
    def test_set_power_state_forcerestart(self, libvirt_mock):
        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.isActive.return_value = True

        self.test_driver.set_power_state(self.uuid, 'ForceRestart')

        domain_mock.reset.assert_called_once_with()

    @mock.patch('libvirt.open', autospec=True)
    def test_set_power_state_nmi(self, libvirt_mock):
        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.isActive.return_value = True

        self.test_driver.set_power_state(self.uuid, 'Nmi')

        domain_mock.injectNMI.assert_called_once_with()

    @mock.patch('libvirt.open', autospec=True)
    def test_power_cycle(self, libvirt_mock):
        with open('sushy_tools/tests/unit/emulator/'
                  'domain_boot_os.xml', 'r') as f:
            data = f.read()

        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.XMLDesc.return_value = data

        with mock.patch.object(
                self.test_driver, 'get_power_state') as gps_mock:
            with mock.patch.object(
                    self.test_driver, 'set_power_state') as sps_mock:
                gps_mock.return_value = 'Off'
                self.test_driver.set_boot_device(self.uuid, 'Cd')

        self.assertFalse(gps_mock.called)
        self.assertFalse(sps_mock.called)

    @mock.patch('libvirt.openReadOnly', autospec=True)
    def test_get_boot_device_os(self, libvirt_mock):
        with open('sushy_tools/tests/unit/emulator/'
                  'domain_boot_os.xml', 'r') as f:
            data = f.read()

        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.XMLDesc.return_value = data

        boot_device = self.test_driver.get_boot_device(self.uuid)

        self.assertEqual('Cd', boot_device)

    @mock.patch('libvirt.open', autospec=True)
    def test_set_boot_device_os(self, libvirt_mock):
        with open('sushy_tools/tests/unit/emulator/'
                  'domain_boot_os.xml', 'r') as f:
            data = f.read()

        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.XMLDesc.return_value = data

        with mock.patch.object(
                self.test_driver, 'get_power_state', return_value='Off'):
            self.test_driver.set_boot_device(self.uuid, 'Hdd')

        conn_mock.defineXML.assert_called_once_with(mock.ANY)

    @mock.patch('libvirt.openReadOnly', autospec=True)
    def test_get_boot_device_disk(self, libvirt_mock):
        with open('sushy_tools/tests/unit/emulator/'
                  'domain_boot_disk.xml', 'r') as f:
            data = f.read()

        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.XMLDesc.return_value = data

        boot_device = self.test_driver.get_boot_device(self.uuid)

        self.assertEqual('Cd', boot_device)

    @mock.patch('libvirt.open', autospec=True)
    def test_set_boot_device_disk(self, libvirt_mock):
        with open('sushy_tools/tests/unit/emulator/'
                  'domain_boot_disk.xml', 'r') as f:
            data = f.read()

        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.XMLDesc.return_value = data

        with mock.patch.object(
                self.test_driver, 'get_power_state', return_value='Off'):
            self.test_driver.set_boot_device(self.uuid, 'Hdd')

        conn_mock.defineXML.assert_called_once_with(mock.ANY)

        expected = '<disk device="disk" type="file">\n      <source ' \
                   'file="/home/user/fedora.img" />\n      <target ' \
                   'dev="hda" />\n    <boot order="1" /></disk>\n'

        # NOTE(rpittau): starting from Python 3.8 the tostring() function
        # preserves the attribute order specified by the user.
        if sys.version_info[1] >= 8:
            expected = '<disk type="file" device="disk">\n      <source ' \
                       'file="/home/user/fedora.img" />\n      <target ' \
                       'dev="hda" />\n    <boot order="1" /></disk>\n'

        self.assertIn(expected, conn_mock.defineXML.call_args[0][0])

    @mock.patch('libvirt.openReadOnly', autospec=True)
    def test_get_boot_device_network(self, libvirt_mock):
        with open('sushy_tools/tests/unit/emulator/'
                  'domain_boot_network.xml', 'r') as f:
            data = f.read()

        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.XMLDesc.return_value = data

        boot_device = self.test_driver.get_boot_device(self.uuid)

        self.assertEqual('Pxe', boot_device)

    @mock.patch('libvirt.open', autospec=True)
    def test_set_boot_device_network(self, libvirt_mock):
        with open('sushy_tools/tests/unit/emulator/'
                  'domain_boot_network.xml', 'r') as f:
            data = f.read()

        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.XMLDesc.return_value = data

        with mock.patch.object(
                self.test_driver, 'get_power_state', return_value='Off'):
            self.test_driver.set_boot_device(self.uuid, 'Pxe')

        conn_mock.defineXML.assert_called_once_with(mock.ANY)

        expected = '<interface type="direct">\n      <mac address=' \
                   '"52:54:00:da:ac:54" />\n      <source dev="tap-' \
                   'node-2i1" mode="vepa" />\n      <model type="vir' \
                   'tio" />\n      <address bus="0x01" domain="0x' \
                   '0000" function="0x0" slot="0x01" type="pci" />'

        # NOTE(rpittau): starting from Python 3.8 the tostring() function
        # preserves the attribute order specified by the user.
        if sys.version_info[1] >= 8:
            expected = '<interface type="direct">\n      <mac address=' \
                       '"52:54:00:da:ac:54" />\n      <source dev="tap-' \
                       'node-2i1" mode="vepa" />\n      <model type="vir' \
                       'tio" />\n      <address type="pci" domain="0x0000" ' \
                       'bus="0x01" slot="0x01" function="0x0" />'

        self.assertIn(expected, conn_mock.defineXML.call_args[0][0])

    @mock.patch('libvirt.openReadOnly', autospec=True)
    def test_get_boot_mode(self, libvirt_mock):
        with open('sushy_tools/tests/unit/emulator/domain.xml', 'r') as f:
            data = f.read()

        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.XMLDesc.return_value = data

        boot_mode = self.test_driver.get_boot_mode(self.uuid)

        self.assertEqual('Legacy', boot_mode)

    @mock.patch('libvirt.open', autospec=True)
    @mock.patch('libvirt.openReadOnly', autospec=True)
    def test_set_boot_mode(self, libvirt_mock, libvirt_rw_mock):
        with open('sushy_tools/tests/unit/emulator/domain.xml', 'r') as f:
            data = f.read()

        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.XMLDesc.return_value = data

        with mock.patch.object(
                self.test_driver, 'get_power_state', return_value='Off'):
            self.test_driver.set_boot_mode(self.uuid, 'UEFI')

        conn_mock = libvirt_rw_mock.return_value
        conn_mock.defineXML.assert_called_once_with(mock.ANY)

    @mock.patch('libvirt.open', autospec=True)
    @mock.patch('libvirt.openReadOnly', autospec=True)
    def test_set_boot_mode_known_loader(self, libvirt_mock, libvirt_rw_mock):
        with open('sushy_tools/tests/unit/emulator/domain.xml', 'r') as f:
            data = f.read()

        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.XMLDesc.return_value = data

        with mock.patch.object(
                self.test_driver, 'get_power_state', return_value='Off'):
            self.test_driver.set_boot_mode(self.uuid, 'UEFI')

        conn_mock = libvirt_rw_mock.return_value
        xml_document = conn_mock.defineXML.call_args[0][0]
        tree = ET.fromstring(xml_document)
        os_element = tree.find('os')
        loader_element = os_element.find('loader')
        self.assertEqual(
            'pflash', loader_element.get('type'))
        self.assertEqual(
            '/usr/share/OVMF/OVMF_CODE.fd',
            loader_element.text)

    @mock.patch('libvirt.open', autospec=True)
    @mock.patch('libvirt.openReadOnly', autospec=True)
    def test_set_boot_mode_unknown_loader_path(
            self, libvirt_mock, libvirt_rw_mock):
        with open('sushy_tools/tests/unit/emulator/domain.xml', 'r') as f:
            data = f.read()

        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.XMLDesc.return_value = data

        with mock.patch.dict(
                self.test_driver.KNOWN_BOOT_LOADERS, {}, clear=True):
            with mock.patch.object(
                    self.test_driver, 'get_power_state', return_value='Off'):
                self.assertRaises(
                    error.FishyError, self.test_driver.set_boot_mode,
                    self.uuid, 'Uefi')

    @mock.patch('libvirt.open', autospec=True)
    @mock.patch('libvirt.openReadOnly', autospec=True)
    def test_set_boot_mode_no_os(self, libvirt_mock, libvirt_rw_mock):
        with open('sushy_tools/tests/unit/emulator/domain.xml', 'r') as f:
            data = f.read()

        tree = ET.fromstring(data)
        os_element = tree.find('os')
        tree.remove(os_element)

        data = ET.tostring(tree)

        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.XMLDesc.return_value = data

        with mock.patch.object(
                self.test_driver, 'get_power_state', return_value='Off'):
            self.assertRaises(
                error.FishyError, self.test_driver.set_boot_mode,
                self.uuid, 'Uefi')

    @mock.patch('libvirt.open', autospec=True)
    @mock.patch('libvirt.openReadOnly', autospec=True)
    def test_set_boot_mode_many_loaders(self, libvirt_mock, libvirt_rw_mock):
        with open('sushy_tools/tests/unit/emulator/domain.xml', 'r') as f:
            data = f.read()

        tree = ET.fromstring(data)
        os_element = tree.find('os')
        ET.SubElement(os_element, 'loader')

        data = ET.tostring(tree)

        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.XMLDesc.return_value = data

        with mock.patch.object(
                self.test_driver, 'get_power_state', return_value='Off'):
            self.assertRaises(
                error.FishyError, self.test_driver.set_boot_mode,
                self.uuid, 'Uefi')

    @mock.patch('libvirt.open', autospec=True)
    @mock.patch('libvirt.openReadOnly', autospec=True)
    def test_set_boot_mode_many_os(self, libvirt_mock, libvirt_rw_mock):
        with open('sushy_tools/tests/unit/emulator/domain.xml', 'r') as f:
            data = f.read()

        tree = ET.fromstring(data)
        ET.SubElement(tree, 'os')
        ET.SubElement(tree, 'os')

        data = ET.tostring(tree)

        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.XMLDesc.return_value = data

        with mock.patch.object(
                self.test_driver, 'get_power_state', return_value='Off'):
            self.assertRaises(
                error.FishyError, self.test_driver.set_boot_mode,
                self.uuid, 'Uefi')

    @mock.patch('libvirt.open', autospec=True)
    @mock.patch('libvirt.openReadOnly', autospec=True)
    def test_set_boot_mode_no_type(self, libvirt_mock, libvirt_rw_mock):
        with open('sushy_tools/tests/unit/emulator/domain.xml', 'r') as f:
            data = f.read()

        tree = ET.fromstring(data)
        os_element = tree.find('os')
        type_element = os_element.find('type')
        os_element.remove(type_element)

        data = ET.tostring(tree)

        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.XMLDesc.return_value = data

        with mock.patch.object(
                self.test_driver, 'get_power_state', return_value='Off'):
            self.test_driver.set_boot_mode(self.uuid, 'UEFI')

        conn_mock = libvirt_rw_mock.return_value
        xml_document = conn_mock.defineXML.call_args[0][0]
        tree = ET.fromstring(xml_document)
        os_element = tree.find('os')
        # NOTE(etingof): should enforce default loader
        self.assertIsNone(os_element.find('loader'))

    @mock.patch('libvirt.openReadOnly', autospec=True)
    def test_get_boot_image(self, libvirt_mock):
        with open('sushy_tools/tests/unit/emulator/domain.xml', 'r') as f:
            data = f.read()

        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.XMLDesc.return_value = data

        image_info = self.test_driver.get_boot_image(self.uuid, 'Cd')

        expected = '/home/user/boot.iso', False, False

        self.assertEqual(expected, image_info)

    @mock.patch('sushy_tools.emulator.resources.systems.libvirtdriver'
                '.os.stat', autospec=True)
    @mock.patch('sushy_tools.emulator.resources.systems.libvirtdriver'
                '.open')
    @mock.patch('libvirt.open', autospec=True)
    @mock.patch('libvirt.openReadOnly', autospec=True)
    def test_set_boot_image(self, libvirt_mock, libvirt_rw_mock,
                            open_mock, stat_mock):
        with open('sushy_tools/tests/unit/emulator/domain.xml', 'r') as f:
            data = f.read()

        conn_mock = libvirt_rw_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.XMLDesc.return_value = data

        pool_mock = conn_mock.storagePoolLookupByName.return_value

        with open('sushy_tools/tests/unit/emulator/pool.xml', 'r') as f:
            data = f.read()

        pool_mock.XMLDesc.return_value = data

        with mock.patch.object(
                self.test_driver, 'get_power_state', return_value='Off'):
            with mock.patch.object(
                    self.test_driver, 'get_boot_device', return_value=None):

                self.test_driver.set_boot_image(
                    self.uuid, 'Cd', '/tmp/image.iso')

        conn_mock = libvirt_rw_mock.return_value
        pool_mock.listAllVolumes.assert_called_once_with()
        stat_mock.assert_called_once_with('/tmp/image.iso')
        pool_mock.createXML.assert_called_once_with(mock.ANY)

        volume_mock = pool_mock.createXML.return_value
        volume_mock.upload.assert_called_once_with(mock.ANY, 0, mock.ANY)

        expected_disk = ('<disk device="cdrom" type="file">'
                         '<target bus="ide" dev="hdc" />'
                         '<address bus="0" controller="0" '
                         'target="0" type="drive" unit="0" />')

        # NOTE(rpittau): starting from Python 3.8 the tostring() function
        # preserves the attribute order specified by the user.
        if sys.version_info[1] >= 8:
            expected_disk = ('<disk type="file" device="cdrom">'
                             '<target dev="hdc" bus="ide" />'
                             '<address type="drive" controller="0"'
                             ' bus="0" target="0" unit="0" />')
        self.assertEqual(1, conn_mock.defineXML.call_count)
        self.assertIn(expected_disk, conn_mock.defineXML.call_args[0][0])

    @mock.patch('sushy_tools.emulator.resources.systems.libvirtdriver'
                '.os.stat', autospec=True)
    @mock.patch('sushy_tools.emulator.resources.systems.libvirtdriver'
                '.open')
    @mock.patch('libvirt.open', autospec=True)
    @mock.patch('libvirt.openReadOnly', autospec=True)
    def test_set_boot_image_sata(self, libvirt_mock, libvirt_rw_mock,
                                 open_mock, stat_mock):
        with open('sushy_tools/tests/unit/emulator/domain-sata.xml', 'r') as f:
            data = f.read()

        conn_mock = libvirt_rw_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.XMLDesc.return_value = data

        pool_mock = conn_mock.storagePoolLookupByName.return_value

        with open('sushy_tools/tests/unit/emulator/pool.xml', 'r') as f:
            data = f.read()

        pool_mock.XMLDesc.return_value = data

        with mock.patch.object(
                self.test_driver, 'get_power_state', return_value='Off'):
            with mock.patch.object(
                    self.test_driver, 'get_boot_device', return_value=None):

                self.test_driver.set_boot_image(
                    self.uuid, 'Cd', '/tmp/image.iso')

        conn_mock = libvirt_rw_mock.return_value
        pool_mock.listAllVolumes.assert_called_once_with()
        stat_mock.assert_called_once_with('/tmp/image.iso')
        pool_mock.createXML.assert_called_once_with(mock.ANY)

        volume_mock = pool_mock.createXML.return_value
        volume_mock.upload.assert_called_once_with(mock.ANY, 0, mock.ANY)

        expected_disk = ('<disk device="cdrom" type="file">'
                         '<target bus="sata" dev="sdc" />'
                         '<address bus="0" controller="0" '
                         'target="0" type="drive" unit="1" />')

        # NOTE(rpittau): starting from Python 3.8 the tostring() function
        # preserves the attribute order specified by the user.
        if sys.version_info[1] >= 8:
            expected_disk = ('<disk type="file" device="cdrom">'
                             '<target dev="sdc" bus="sata" />'
                             '<address type="drive" controller="0"'
                             ' bus="0" target="0" unit="1" />')
        self.assertEqual(1, conn_mock.defineXML.call_count)
        self.assertIn(expected_disk, conn_mock.defineXML.call_args[0][0])

    @mock.patch('sushy_tools.emulator.resources.systems.libvirtdriver'
                '.os.stat', autospec=True)
    @mock.patch('sushy_tools.emulator.resources.systems.libvirtdriver'
                '.open')
    @mock.patch('libvirt.open', autospec=True)
    @mock.patch('libvirt.openReadOnly', autospec=True)
    def test_set_boot_image_scsi(self, libvirt_mock, libvirt_rw_mock,
                                 open_mock, stat_mock):
        with open('sushy_tools/tests/unit/emulator/domain-scsi.xml', 'r') as f:
            data = f.read()

        conn_mock = libvirt_rw_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.XMLDesc.return_value = data

        pool_mock = conn_mock.storagePoolLookupByName.return_value

        with open('sushy_tools/tests/unit/emulator/pool.xml', 'r') as f:
            data = f.read()

        pool_mock.XMLDesc.return_value = data

        with mock.patch.object(
                self.test_driver, 'get_power_state', return_value='Off'):
            with mock.patch.object(
                    self.test_driver, 'get_boot_device', return_value=None):

                self.test_driver.set_boot_image(
                    self.uuid, 'Cd', '/tmp/image.iso')

        conn_mock = libvirt_rw_mock.return_value
        pool_mock.listAllVolumes.assert_called_once_with()
        stat_mock.assert_called_once_with('/tmp/image.iso')
        pool_mock.createXML.assert_called_once_with(mock.ANY)

        volume_mock = pool_mock.createXML.return_value
        volume_mock.upload.assert_called_once_with(mock.ANY, 0, mock.ANY)

        expected_disk = ('<disk device="cdrom" type="file">'
                         '<target bus="scsi" dev="sdc" />'
                         '<address bus="0" controller="0" '
                         'target="0" type="drive" unit="1" />')

        # NOTE(rpittau): starting from Python 3.8 the tostring() function
        # preserves the attribute order specified by the user.
        if sys.version_info[1] >= 8:
            expected_disk = ('<disk type="file" device="cdrom">'
                             '<target dev="sdc" bus="scsi" />'
                             '<address type="drive" controller="0"'
                             ' bus="0" target="0" unit="1" />')

        self.assertEqual(1, conn_mock.defineXML.call_count)
        self.assertIn(expected_disk, conn_mock.defineXML.call_args[0][0])

    @mock.patch('libvirt.open', autospec=True)
    @mock.patch('libvirt.openReadOnly', autospec=True)
    @mock.patch(
        'sushy_tools.emulator.resources.systems.libvirtdriver.LibvirtDriver'
        '.get_power_state', new=mock.MagicMock(return_value='Off'))
    @mock.patch(
        'sushy_tools.emulator.resources.systems.libvirtdriver.LibvirtDriver'
        '.get_boot_device', return_value='Cd')
    @mock.patch(
        'sushy_tools.emulator.resources.systems.libvirtdriver.LibvirtDriver'
        '.set_boot_device')
    @mock.patch(
        'sushy_tools.emulator.resources.systems.libvirtdriver.LibvirtDriver'
        '._add_boot_image', new=mock.MagicMock())
    @mock.patch(
        'sushy_tools.emulator.resources.systems.libvirtdriver.LibvirtDriver'
        '._remove_boot_images', new=mock.MagicMock())
    def test_set_boot_image_restore_boot_device(
            self, sbd_mock, gbd_mock, libvirt_mock, libvirt_rw_mock):

        with open('sushy_tools/tests/unit/emulator/domain.xml', 'r') as f:
            data = f.read()

        conn_mock = libvirt_rw_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.XMLDesc.return_value = data

        self.test_driver.set_boot_image(
            self.uuid, 'Cd', '/tmp/image.iso')

        gbd_mock.assert_called_once_with(self.uuid)
        sbd_mock.assert_called_once_with(self.uuid, 'Cd')

    @mock.patch('libvirt.openReadOnly', autospec=True)
    def test_get_total_memory(self, libvirt_mock):
        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.maxMemory.return_value = 1024 * 1024

        memory = self.test_driver.get_total_memory(self.uuid)

        self.assertEqual(1, memory)

    @mock.patch('libvirt.openReadOnly', autospec=True)
    def test_get_total_cpus(self, libvirt_mock):
        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.isActive.return_value = True
        domain_mock.XMLDesc.return_value = b'<empty/>'
        domain_mock.maxVcpus.return_value = 2

        cpus = self.test_driver.get_total_cpus(self.uuid)

        self.assertEqual(2, cpus)

    @mock.patch('libvirt.open', autospec=True)
    def test_get_bios(self, libvirt_mock):
        with open('sushy_tools/tests/unit/emulator/domain.xml') as f:
            domain_xml = f.read()

        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.XMLDesc.return_value = domain_xml

        bios_attributes = self.test_driver.get_bios(self.uuid)
        self.assertEqual(LibvirtDriver.DEFAULT_BIOS_ATTRIBUTES,
                         bios_attributes)
        conn_mock.defineXML.assert_called_once_with(mock.ANY)

    @mock.patch('libvirt.open', autospec=True)
    def test_get_bios_existing(self, libvirt_mock):
        with open('sushy_tools/tests/unit/emulator/domain_bios.xml') as f:
            domain_xml = f.read()

        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.XMLDesc.return_value = domain_xml

        bios_attributes = self.test_driver.get_bios(self.uuid)
        self.assertEqual({"BootMode": "Bios",
                          "EmbeddedSata": "Raid",
                          "NicBoot1": "NetworkBoot",
                          "ProcTurboMode": "Disabled"},
                         bios_attributes)
        conn_mock.defineXML.assert_not_called()

    @mock.patch('libvirt.open', autospec=True)
    def test_set_bios(self, libvirt_mock):
        with open('sushy_tools/tests/unit/emulator/domain_bios.xml') as f:
            domain_xml = f.read()

        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.XMLDesc.return_value = domain_xml

        with mock.patch.object(
                self.test_driver, 'get_power_state', return_value='Off'):
            self.test_driver.set_bios(
                self.uuid, {"BootMode": "Uefi",
                            "ProcTurboMode": "Enabled"})

        conn_mock.defineXML.assert_called_once_with(mock.ANY)

    @mock.patch('libvirt.open', autospec=True)
    def test_reset_bios(self, libvirt_mock):
        with open('sushy_tools/tests/unit/emulator/domain_bios.xml') as f:
            domain_xml = f.read()

        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.XMLDesc.return_value = domain_xml

        with mock.patch.object(
                self.test_driver, 'get_power_state', return_value='Off'):
            self.test_driver.reset_bios(self.uuid)

        conn_mock.defineXML.assert_called_once_with(mock.ANY)

    def test__process_bios_attributes_get_default(self):
        with open('sushy_tools/tests/unit/emulator/domain.xml') as f:
            domain_xml = f.read()

        result = self.test_driver._process_bios_attributes(domain_xml)
        self.assertTrue(result.attributes_written)
        self.assertEqual(LibvirtDriver.DEFAULT_BIOS_ATTRIBUTES,
                         result.bios_attributes)
        self._assert_bios_xml(result.tree)

    def test__process_bios_attributes_get_default_metadata_exists(self):
        with open('sushy_tools/tests/unit/emulator/'
                  'domain_metadata.xml') as f:
            domain_xml = f.read()

        result = self.test_driver._process_bios_attributes(domain_xml)
        self.assertTrue(result.attributes_written)
        self.assertEqual(LibvirtDriver.DEFAULT_BIOS_ATTRIBUTES,
                         result.bios_attributes)
        self._assert_bios_xml(result.tree)

    def test__process_bios_attributes_get_existing(self):
        with open('sushy_tools/tests/unit/emulator/domain_bios.xml') as f:
            domain_xml = f.read()

        result = self.test_driver._process_bios_attributes(domain_xml)
        self.assertFalse(result.attributes_written)
        self.assertEqual({"BootMode": "Bios",
                          "EmbeddedSata": "Raid",
                          "NicBoot1": "NetworkBoot",
                          "ProcTurboMode": "Disabled"},
                         result.bios_attributes)
        self._assert_bios_xml(result.tree)

    def test__process_bios_attributes_update(self):
        with open('sushy_tools/tests/unit/emulator/domain_bios.xml') as f:
            domain_xml = f.read()
        result = self.test_driver._process_bios_attributes(
            domain_xml,
            {"BootMode": "Uefi",
             "ProcTurboMode": "Enabled"},
            True)
        self.assertTrue(result.attributes_written)
        self.assertEqual({"BootMode": "Uefi",
                         "ProcTurboMode": "Enabled"},
                         result.bios_attributes)
        self._assert_bios_xml(result.tree)

    def _assert_bios_xml(self, tree):
        ns = {'sushy': 'http://openstack.org/xmlns/libvirt/sushy'}
        self.assertIsNotNone(tree.find('metadata')
                             .find('sushy:bios', ns)
                             .find('sushy:attributes', ns))

    @mock.patch('libvirt.open', autospec=True)
    def test__process_bios_error(self, libvirt_mock):
        with open('sushy_tools/tests/unit/emulator/domain.xml') as f:
            domain_xml = f.read()

        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.XMLDesc.return_value = domain_xml
        conn_mock.defineXML.side_effect = libvirt.libvirtError(
            'because I can')

        self.assertRaises(error.FishyError,
                          self.test_driver._process_bios,
                          'xxx-yyy-zzz',
                          {"BootMode": "Uefi",
                           "ProcTurboMode": "Enabled"})

    @mock.patch('libvirt.openReadOnly', autospec=True)
    def test_get_nics(self, libvirt_mock):
        with open('sushy_tools/tests/unit/emulator/domain_nics.xml') as f:
            domain_xml = f.read()

        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.XMLDesc.return_value = domain_xml

        nics = self.test_driver.get_nics(self.uuid)
        self.assertEqual([{'id': '00:11:22:33:44:55',
                           'mac': '00:11:22:33:44:55'},
                          {'id': '52:54:00:4e:5d:37',
                           'mac': '52:54:00:4e:5d:37'}],
                         sorted(nics, key=lambda k: k['id']))

    @mock.patch('libvirt.openReadOnly', autospec=True)
    def test_get_nics_empty(self, libvirt_mock):
        with open('sushy_tools/tests/unit/emulator/domain.xml') as f:
            domain_xml = f.read()

        conn_mock = libvirt_mock.return_value
        domain_mock = conn_mock.lookupByUUID.return_value
        domain_mock.XMLDesc.return_value = domain_xml

        nics = self.test_driver.get_nics(self.uuid)
        self.assertEqual([], nics)

    @mock.patch('libvirt.openReadOnly', autospec=True)
    def test_get_simple_storage_collection(self, libvirt_mock):
        with open('sushy_tools/tests/unit/emulator/'
                  'domain_simple_storage.xml', 'r') as f:
            data = f.read()

        conn_mock = libvirt_mock.return_value
        dom_mock = conn_mock.lookupByUUID.return_value
        dom_mock.XMLDesc.return_value = data
        vol_mock = conn_mock.storageVolLookupByPath.return_value
        vol_mock.name.side_effect = ['testVM1.img', 'testVol1.img', 'sdb1']
        vol_mock.info.side_effect = [['testVM1.img', 100000],
                                     ['testVol1.img', 200000],
                                     ['sdb1', 150000]]
        pool_mock = conn_mock.storagePoolLookupByName.return_value
        pvol_mock = pool_mock.storageVolLookupByName.return_value
        pvol_mock.name.return_value = 'blk-pool0-vol0'
        pvol_mock.info.return_value = ['volType', 300000]

        simple_storage_response = (self.test_driver
                                   .get_simple_storage_collection(self.uuid))

        simple_storage_expected = {
            'virtio': {
                'Id': 'virtio',
                'Name': 'virtio',
                'DeviceList': [
                    {
                        'Name': 'testVM1.img',
                        'CapacityBytes': 100000
                    },
                    {
                        'Name': 'sdb1',
                        'CapacityBytes': 150000
                    }
                ]
            },
            'ide': {
                'Id': 'ide',
                'Name': 'ide',
                'DeviceList': [
                    {
                        'Name': 'testVol1.img',
                        'CapacityBytes': 200000
                    },
                    {
                        'Name': 'blk-pool0-vol0',
                        'CapacityBytes': 300000
                    }
                ]
            }
        }

        self.assertEqual(simple_storage_response, simple_storage_expected)

    @mock.patch('libvirt.openReadOnly', autospec=True)
    def test_get_simple_storage_collection_empty(self, libvirt_mock):
        with open('sushy_tools/tests/unit/emulator/domain.xml') as f:
            domain_xml = f.read()

        conn_mock = libvirt_mock.return_value
        dom_mock = conn_mock.lookupByUUID.return_value
        dom_mock.XMLDesc.return_value = domain_xml
        vol_mock = conn_mock.storageVolLookupByPath.return_value
        vol_mock.name.side_effect = ['testVM1.img', 'testVol1.img', 'sdb1']
        vol_mock.info.side_effect = [['testType1', 100000],
                                     ['testType2', 200000],
                                     ['testType1', 150000]]
        pool_mock = conn_mock.storagePoolLookupByName.return_value
        pvol_mock = pool_mock.storageVolLookupByName.return_value
        pvol_mock.name.return_value = 'blk-pool0-vol0'
        pvol_mock.info.return_value = ['volType', 300000]

        simple_storage_response = (self.test_driver
                                   .get_simple_storage_collection(self.uuid))

        self.assertEqual({}, simple_storage_response)

    @mock.patch('libvirt.open', autospec=True)
    def test_find_or_create_storage_volume(self, libvirt_mock):
        conn_mock = libvirt_mock.return_value
        vol_data = {
            "libvirtVolName": "123456",
            "Id": "1",
            "Name": "Sample Vol",
            "CapacityBytes": 12345,
            "VolumeType": "Mirrored"
        }

        pool_mock = conn_mock.storagePoolLookupByName.return_value
        with open('sushy_tools/tests/unit/emulator/pool.xml', 'r') as f:
            data = f.read()
        pool_mock.storageVolLookupByName.side_effect = libvirt.libvirtError(
            'Storage volume not found')
        pool_mock.XMLDesc.return_value = data

        self.test_driver.find_or_create_storage_volume(vol_data)
        pool_mock.createXML.assert_called_once_with(mock.ANY)



# Copyright 2019 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslotest import base
from six.moves import mock

from sushy_tools.emulator.resources.storage.staticdriver import StaticDriver


class StaticDriverTestCase(base.BaseTestCase):
    UUID = "da69abcc-dae0-4913-9a7b-d344043097c0"
    STORAGE_COL = [
        {
            "Id": "1",
            "Name": "Local Storage Controller",
            "StorageControllers": [
                {
                    "MemberId": "0",
                    "Name": "Contoso Integrated RAID",
                    "SpeedGbps": 12
                }
            ]
        }
    ]

    CONFIG = {
        'SUSHY_EMULATOR_STORAGE': {
            UUID: STORAGE_COL
        }
    }

    def test_get_storage_col(self):
        test_driver = StaticDriver.initialize(
            self.CONFIG, mock.MagicMock())()
        stg_col = test_driver.get_storage_col(self.UUID)
        self.assertEqual(self.STORAGE_COL, stg_col)

    def test_get_all_storage(self):
        test_driver = StaticDriver.initialize(
            self.CONFIG, mock.MagicMock())()
        stg = test_driver.get_all_storage()
        self.assertEqual([(self.UUID, '1')], stg)



# Copyright 2018 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from unittest import mock

from oslotest import base

from sushy_tools.emulator import memoize


class MemoizeTestCase(base.BaseTestCase):

    def test_local_cache(self):

        class Driver(object):
            call_count = 0

            @memoize.memoize()
            def fun(self, *args, **kwargs):
                self.call_count += 1
                return args, kwargs

        driver = Driver()

        result = driver.fun(1, x=2)
        self.assertEqual(((1,), {'x': 2}), result)

        result = driver.fun(1, x=2)
        self.assertEqual(((1,), {'x': 2}), result)

        # Due to memoization, expect just one call
        self.assertEqual(1, driver.call_count)

        driver = Driver()

        result = driver.fun(1, x=2)
        self.assertEqual(((1,), {'x': 2}), result)

        # Due to volatile cache, expect one more call
        self.assertEqual(1, driver.call_count)

    def test_external_cache(self):

        permanent_cache = {}

        class Driver(object):
            call_count = 0

            @memoize.memoize(permanent_cache=permanent_cache)
            def fun(self, *args, **kwargs):
                self.call_count += 1
                return args, kwargs

        driver = Driver()

        result = driver.fun(1, x=2)
        self.assertEqual(((1,), {'x': 2}), result)

        result = driver.fun(1, x=2)
        self.assertEqual(((1,), {'x': 2}), result)

        # Due to memoization, expect just one call
        self.assertEqual(1, driver.call_count)

        driver = Driver()

        result = driver.fun(1, x=2)
        self.assertEqual(((1,), {'x': 2}), result)

        # Due to permanent cache, expect no more calls
        self.assertEqual(0, driver.call_count)


class PersistentDictTestCase(base.BaseTestCase):

    @mock.patch.object(memoize, 'sqlite3', autospec=True)
    def test_make_permanent(self, mock_sqlite3):
        pd = memoize.PersistentDict()

        pd.make_permanent('/', 'file')

        mock_sqlite3.connect.assert_called_with('/file.sqlite')

    @mock.patch.object(memoize, 'pickle', autospec=True)
    def test_encode(self, mock_pickle):
        pd = memoize.PersistentDict()

        pd.encode({1: '2'})

        mock_pickle.dumps.assert_called_once_with({1: '2'})

    @mock.patch.object(memoize, 'pickle', autospec=True)
    def test_decode(self, mock_pickle):
        pd = memoize.PersistentDict()

        pd.decode('blob')

        mock_pickle.loads.assert_called_once_with('blob')

    @mock.patch.object(memoize, 'pickle', autospec=True)
    @mock.patch.object(memoize, 'sqlite3', autospec=True)
    def test___getitem__(self, mock_sqlite3, mock_pickle):
        pd = memoize.PersistentDict()
        pd.make_permanent('/', 'file')

        mock_pickle.dumps.return_value = 'pickled-key'
        mock_connection = mock_sqlite3.connect.return_value
        mock_cursor = mock_connection.cursor.return_value
        mock_cursor.fetchone.return_value = ['pickled-value']

        pd[1]

        mock_cursor.execute.assert_called_with(
            'select value from cache where key=?', ('pickled-key',))
        mock_pickle.loads.assert_called_once_with('pickled-value')

    @mock.patch.object(memoize, 'pickle', autospec=True)
    @mock.patch.object(memoize, 'sqlite3', autospec=True)
    def test___setitem__(self, mock_sqlite3, mock_pickle):
        pd = memoize.PersistentDict()
        pd.make_permanent('/', 'file')

        mock_pickle.dumps.side_effect = [
            'pickled-key', 'pickled-value']

        mock_connection = mock_sqlite3.connect.return_value
        mock_cursor = mock_connection.cursor.return_value

        pd[1] = 2

        mock_cursor.execute.assert_called_with(
            'insert or replace into cache values (?, ?)',
            ('pickled-key', 'pickled-value'))

    @mock.patch.object(memoize, 'pickle', autospec=True)
    @mock.patch.object(memoize, 'sqlite3', autospec=True)
    def test___delitem__(self, mock_sqlite3, mock_pickle):
        pd = memoize.PersistentDict()
        pd.make_permanent('/', 'file')

        mock_pickle.dumps.return_value = 'pickled-key'
        mock_connection = mock_sqlite3.connect.return_value
        mock_cursor = mock_connection.cursor.return_value

        del pd[1]

        mock_cursor.execute.assert_called_with(
            'delete from cache where key=?', ('pickled-key',))

    @mock.patch.object(memoize, 'pickle', autospec=True)
    @mock.patch.object(memoize, 'sqlite3', autospec=True)
    def test___iter__(self, mock_sqlite3, mock_pickle):
        pd = memoize.PersistentDict()
        pd.make_permanent('/', 'file')

        mock_pickle.dumps.return_value = 'pickled-key'
        mock_connection = mock_sqlite3.connect.return_value
        mock_cursor = mock_connection.cursor.return_value
        mock_cursor.fetchall.return_value = [['pickled-key']]

        for x in pd:
            x += x

        mock_cursor.execute.assert_called_with('select key from cache')
        mock_pickle.loads.assert_called_once_with('pickled-key')

    @mock.patch.object(memoize, 'pickle', autospec=True)
    @mock.patch.object(memoize, 'sqlite3', autospec=True)
    def test___len__(self, mock_sqlite3, mock_pickle):
        pd = memoize.PersistentDict()
        pd.make_permanent('/', 'file')

        mock_connection = mock_sqlite3.connect.return_value
        mock_cursor = mock_connection.cursor.return_value

        expected = 1

        mock_cursor.fetchone.return_value = [expected]

        self.assertEqual(expected, len(pd))

        mock_cursor.execute.assert_called_with('select count(*) from cache')



# Copyright 2019 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import uuid

from oslotest import base
from six.moves import mock

from sushy_tools.emulator.resources.managers.staticdriver import StaticDriver
from sushy_tools import error


class StaticDriverTestCase(base.BaseTestCase):

    def setUp(self):
        self.managers = [
            {
                "Id": "BMC",
                "Name": "The manager",
                "UUID": "58893887-8974-2487-2389-841168418919",
                "ServiceEntryPointUUID": "92384634-2938-2342-8820-489239905423"
            }
        ]

        self.identity = self.managers[0]['Id']
        self.uuid = self.managers[0]['UUID']
        self.name = self.managers[0]['Name']

        test_driver = StaticDriver.initialize(
            {'SUSHY_EMULATOR_MANAGERS': self.managers},
            mock.MagicMock())

        self.test_driver = test_driver()

        super(StaticDriverTestCase, self).setUp()

    def test__get_manager_by_id(self):
        self.assertRaises(
            error.AliasAccessError, self.test_driver._get_manager,
            self.identity)

    def test__get_manager_by_name(self):
        self.assertRaises(
            error.AliasAccessError, self.test_driver._get_manager, self.name)

    def test__get_manager_by_uuid(self):
        domain_id = uuid.UUID(self.uuid)
        manager = self.test_driver._get_manager(str(domain_id))
        self.assertEqual(
            self.managers[0], manager)

    def test_uuid_ok(self):
        self.assertEqual(self.uuid, self.test_driver.uuid(self.uuid))

    def test_uuid_fail(self):
        self.assertRaises(error.FishyError, self.test_driver.uuid, 'xxx')

    def test_name_ok(self):
        self.assertRaises(error.AliasAccessError,
                          self.test_driver.name, self.name)

    def test_name_fail(self):
        self.assertRaises(error.FishyError, self.test_driver.name, 'xxx')

    def test_managers(self):
        managers = self.test_driver.managers
        self.assertEqual([self.uuid], managers)



# Copyright 2019 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from sushy_tools.emulator.resources.base import DriverBase
from sushy_tools import error


class StaticDriver(DriverBase):
    """Redfish storage drives backed by configuration file"""

    @classmethod
    def initialize(cls, config, logger, *args, **kwargs):
        cls._config = config
        cls._logger = logger

        cls._drives = cls._config.get('SUSHY_EMULATOR_DRIVES', {})

        return cls

    @property
    def driver(self):
        """Return human-friendly driver information

        :returns: driver information as `str`
        """
        return '<static-drives>'

    def get_drives(self, identity, storage_id):
        try:
            uu_identity = str(uuid.UUID(identity))

            return self._drives[(uu_identity, storage_id)]

        except (ValueError, KeyError):
            msg = ('Error finding drive for System UUID "%s" and Storage ID '
                   '"%s"', identity, storage_id)

            self._logger.debug(msg)

            raise error.FishyError(msg)

    def get_all_drives(self):
        """Return all drives represented as tuples in the following format:

        (System_UUID, Storage_ID, Drive_ID)

        :returns: list of tuples representing the drives
        """
        return [k + (d['Id'],) for k in self._drives
                for d in self._drives[k]]



# Copyright 2019 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


class DriverBase(object):
    """Common base for emulated Redfish resource drivers"""

    @classmethod
    def initialize(cls, config, logger, *args, **kwargs):
        """Initialize class attributes

        Since drivers may need to cache thing short-term. The emulator
        instantiates the driver every time it serves a client query.

        Driver objects can cache whenever it makes sense for the duration
        of a single session. It is guaranteed that the driver object will
        never be reused for any other session.

        The `initialize` method is provided to set up the driver in a way
        that would affect all the subsequent sessions.

        :params config: system configuration dict
        :params logger: system logger object
        :params *args: driver-specific parameters
        :params **kwargs: driver-specific parameters
        :returns: initialized driver class
        """
        return cls



# Copyright 2019 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from sushy_tools.emulator.resources.base import DriverBase
from sushy_tools import error


class StaticDriver(DriverBase):
    """Redfish storage backed by configuration file"""

    @classmethod
    def initialize(cls, config, logger, *args, **kwargs):
        cls._config = config
        cls._logger = logger

        cls._storage = cls._config.get('SUSHY_EMULATOR_STORAGE', {})

        return cls

    @property
    def driver(self):
        """Return human-friendly driver information

        :returns: driver information as `str`
        """
        return '<static-storage>'

    def get_storage_col(self, identity):
        try:
            uu_identity = str(uuid.UUID(identity))

            return self._storage[uu_identity]

        except KeyError:
            msg = ('Error finding storage collection by UUID '
                   '"%(identity)s"' % {'identity': identity})

            self._logger.debug(msg)

            raise error.FishyError(msg)

    def get_all_storage(self):
        """Returns all storage instances represented as tuples in the format:

        (System_ID, Storage_ID)

        :returns: list of tuples representing the storage instances
        """
        return [(k, st["Id"]) for k in self._storage
                for st in self._storage[k]]



# Copyright 2019 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from sushy_tools.emulator import memoize
from sushy_tools.emulator.resources.base import DriverBase


class StaticDriver(DriverBase):
    """Redfish Volumes emulated in libvirt backed by the config file

    Maintains the libvirt volumes in memory.
    """
    @classmethod
    def initialize(cls, config, logger, *args, **kwargs):
        cls._config = config
        cls._logger = logger

        cls._volumes = memoize.PersistentDict()

        if hasattr(cls._volumes, 'make_permanent'):
            cls._volumes.make_permanent(
                cls._config.get('SUSHY_EMULATOR_STATE_DIR'), 'volumes')

        cls._volumes.update(
            cls._config.get('SUSHY_EMULATOR_VOLUMES', {}))

        return cls

    @property
    def driver(self):
        """Return human-friendly driver information

        :returns: driver information as `str`
        """
        return '<static-volumes>'

    def get_volumes_col(self, identity, storage_id):
        try:
            uu_identity = str(uuid.UUID(identity))

            return self._volumes[(uu_identity, storage_id)]

        except (KeyError, ValueError):
            msg = ('Error finding volume collection by System UUID %s '
                   'and Storage ID %s' % (uu_identity, storage_id))
            self._logger.debug(msg)

    def add_volume(self, uu_identity, storage_id, vol):
        if not self._volumes[(uu_identity, storage_id)]:
            self._volumes[(uu_identity, storage_id)] = []

        vol_col = self._volumes[(uu_identity, storage_id)]
        vol_col.append(vol)
        self._volumes.update({(uu_identity, storage_id): vol_col})

    def delete_volume(self, uu_identity, storage_id, vol):
        try:
            vol_col = self._volumes[(uu_identity, storage_id)]
        except KeyError:
            msg = ('Error finding volume collection by System UUID %s '
                   'and Storage ID %s' % (uu_identity, storage_id))
            self._logger.debug(msg)
        else:
            vol_col.remove(vol)
            self._volumes.update({(uu_identity, storage_id): vol_col})



# Copyright 2017 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import argparse
from datetime import datetime
import functools
import json
import os
import ssl
import sys

import flask

from sushy_tools.emulator.resources.chassis import staticdriver as chsdriver
from sushy_tools.emulator.resources.drives import staticdriver as drvdriver
from sushy_tools.emulator.resources.indicators import staticdriver as inddriver
from sushy_tools.emulator.resources.managers import staticdriver as mgrdriver
from sushy_tools.emulator.resources.storage import staticdriver as stgdriver
from sushy_tools.emulator.resources.systems import libvirtdriver
from sushy_tools.emulator.resources.systems import novadriver
from sushy_tools.emulator.resources.vmedia import staticdriver as vmddriver
from sushy_tools.emulator.resources.volumes import staticdriver as voldriver
from sushy_tools import error
from sushy_tools.error import FishyError


class Application(flask.Flask):

    def __init__(self):
        super().__init__(__name__)
        # Turn off strict_slashes on all routes
        self.url_map.strict_slashes = False
        config_file = os.environ.get('SUSHY_EMULATOR_CONFIG')
        if config_file:
            self.config.from_pyfile(config_file)


app = Application()


class Resources(object):

    SYSTEMS = None
    MANAGERS = None
    CHASSIS = None
    INDICATORS = None
    VMEDIA = None
    STORAGE = None
    DRIVES = None
    VOLUMES = None

    def __new__(cls, *args, **kwargs):

        if not cls.SYSTEMS:

            os_cloud = app.config.get('SUSHY_EMULATOR_OS_CLOUD')

            if os_cloud:
                if not novadriver.is_loaded:
                    app.logger.error('Nova driver not loaded')
                    sys.exit(1)

                cls.SYSTEMS = novadriver.OpenStackDriver.initialize(
                    app.config, app.logger, os_cloud)

            else:
                if not libvirtdriver.is_loaded:
                    app.logger.error('libvirt driver not loaded')
                    sys.exit(1)

                libvirt_uri = app.config.get('SUSHY_EMULATOR_LIBVIRT_URI', '')

                cls.SYSTEMS = libvirtdriver.LibvirtDriver.initialize(
                    app.config, app.logger, libvirt_uri)

            app.logger.debug(
                'Initialized system resource backed by %s '
                'driver', cls.SYSTEMS().driver)

        if cls.MANAGERS is None:
            cls.MANAGERS = mgrdriver.StaticDriver.initialize(
                app.config, app.logger)

            app.logger.debug(
                'Initialized manager resource backed by %s '
                'driver', cls.MANAGERS().driver)

        if cls.CHASSIS is None:
            cls.CHASSIS = chsdriver.StaticDriver.initialize(
                app.config, app.logger)

            app.logger.debug(
                'Initialized chassis resource backed by %s '
                'driver', cls.CHASSIS().driver)

        if cls.INDICATORS is None:
            cls.INDICATORS = inddriver.StaticDriver.initialize(
                app.config, app.logger)

            app.logger.debug(
                'Initialized indicators resource backed by %s '
                'driver', cls.INDICATORS().driver)

        if cls.VMEDIA is None:
            cls.VMEDIA = vmddriver.StaticDriver.initialize(
                app.config, app.logger)

            app.logger.debug(
                'Initialized virtual media resource backed by %s '
                'driver', cls.VMEDIA().driver)

        if cls.STORAGE is None:
            cls.STORAGE = stgdriver.StaticDriver.initialize(
                app.config, app.logger)

            app.logger.debug(
                'Initialized storage resource backed by %s '
                'driver', cls.STORAGE().driver)

        if cls.DRIVES is None:
            cls.DRIVES = drvdriver.StaticDriver.initialize(
                app.config, app.logger)

            app.logger.debug(
                'Initialized drive resource backed by %s '
                'driver', cls.DRIVES().driver)

        if cls.VOLUMES is None:
            cls.VOLUMES = voldriver.StaticDriver.initialize(
                app.config, app.logger)

            app.logger.debug(
                'Initialized volumes resource backed by %s '
                'driver', cls.VOLUMES().driver)

        return super(Resources, cls).__new__(cls, *args, **kwargs)

    def __enter__(self):
        self.systems = self.SYSTEMS()
        self.managers = self.MANAGERS()
        self.chassis = self.CHASSIS()
        self.indicators = self.INDICATORS()
        self.vmedia = self.VMEDIA()
        self.storage = self.STORAGE()
        self.drives = self.DRIVES()
        self.volumes = self.VOLUMES()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        del self.systems
        del self.managers
        del self.chassis
        del self.indicators
        del self.vmedia
        del self.storage
        del self.drives
        del self.volumes


def instance_denied(**kwargs):
    deny = True

    try:
        deny = (kwargs['identity'] not in
                app.config['SUSHY_EMULATOR_ALLOWED_INSTANCES'])

    except KeyError:
        deny = False

    finally:
        if deny:
            app.logger.warning('Instance %s access denied',
                               kwargs.get('identity'))

        return deny


def ensure_instance_access(decorated_func):
    @functools.wraps(decorated_func)
    def decorator(*args, **kwargs):
        if instance_denied(**kwargs):
            raise FishyError('Error finding instance')

        return decorated_func(*args, **kwargs)

    return decorator


def returns_json(decorated_func):
    @functools.wraps(decorated_func)
    def decorator(*args, **kwargs):
        response = decorated_func(*args, **kwargs)
        if isinstance(response, flask.Response):
            return response
        if isinstance(response, tuple):
            contents, status = response
        else:
            contents, status = response, 200
        return flask.Response(response=contents, status=status,
                              content_type='application/json')

    return decorator


@app.errorhandler(Exception)
@returns_json
def all_exception_handler(message):
    if isinstance(message, error.AliasAccessError):
        url = flask.url_for(flask.request.endpoint, identity=message.args[0])
        return flask.redirect(url, code=307, Response=flask.Response)
    return flask.render_template('error.json', message=message), 500


@app.route('/redfish/v1/')
@returns_json
def root_resource():
    return flask.render_template('root.json')


@app.route('/redfish/v1/Chassis')
@returns_json
def chassis_collection_resource():
    with Resources() as resources:

        app.logger.debug('Serving chassis list')

        return flask.render_template(
            'chassis_collection.json',
            manager_count=len(resources.chassis.chassis),
            chassis=resources.chassis.chassis)


@app.route('/redfish/v1/Chassis/<identity>', methods=['GET', 'PATCH'])
@returns_json
def chassis_resource(identity):
    with Resources() as resources:

        chassis = resources.chassis

        uuid = chassis.uuid(identity)

        if flask.request.method == 'GET':

            app.logger.debug('Serving resources for chassis "%s"', identity)

            # the first chassis gets all resources
            if uuid == chassis.chassis[0]:
                systems = resources.systems.systems
                managers = resources.managers.managers
                storage = resources.storage.get_all_storage()
                drives = resources.drives.get_all_drives()

            else:
                systems = []
                managers = []
                storage = []
                drives = []

            return flask.render_template(
                'chassis.json',
                identity=identity,
                name=chassis.name(identity),
                uuid=uuid,
                contained_by=None,
                contained_systems=systems,
                contained_managers=managers,
                contained_chassis=[],
                managers=managers[:1],
                indicator_led=resources.indicators.get_indicator_state(
                    uuid),
                storage=storage,
                drives=drives
            )

        elif flask.request.method == 'PATCH':
            indicator_led_state = flask.request.json.get('IndicatorLED')
            if not indicator_led_state:
                return 'PATCH only works for IndicatorLED element', 400

            resources.indicators.set_indicator_state(
                uuid, indicator_led_state)

            app.logger.info('Set indicator LED to "%s" for chassis "%s"',
                            indicator_led_state, identity)

            return '', 204


@app.route('/redfish/v1/Chassis/<identity>/Thermal', methods=['GET'])
@returns_json
def thermal_resource(identity):
    with Resources() as resources:

        chassis = resources.chassis

        uuid = chassis.uuid(identity)

        if flask.request.method != 'GET':
            return

        app.logger.debug(
            'Serving thermal resources for chassis "%s"', identity)

        # the first chassis gets all resources
        if uuid == chassis.chassis[0]:
            systems = resources.systems.systems

        else:
            systems = []

        return flask.render_template(
            'thermal.json',
            chassis=identity,
            systems=systems
        )


@app.route('/redfish/v1/Managers')
@returns_json
def manager_collection_resource():
    with Resources() as resources:

        app.logger.debug('Serving managers list')

        return flask.render_template(
            'manager_collection.json',
            manager_count=len(resources.managers.managers),
            managers=resources.managers.managers)


@app.route('/redfish/v1/Managers/<identity>', methods=['GET'])
@returns_json
def manager_resource(identity):
    if flask.request.method == 'GET':

        with Resources() as resources:

            app.logger.debug('Serving resources for manager "%s"', identity)

            managers = resources.managers

            uuid = managers.uuid(identity)

            # the first manager gets all resources
            if uuid == managers.managers[0]:
                systems = resources.systems.systems
                chassis = resources.chassis.chassis

            else:
                systems = []
                chassis = []

            return flask.render_template(
                'manager.json',
                dateTime=datetime.now().strftime('%Y-%M-%dT%H:%M:%S+00:00'),
                identity=identity,
                name=resources.managers.name(identity),
                uuid=uuid,
                serviceEntryPointUUID=resources.managers.uuid(identity),
                systems=systems,
                chassis=chassis
            )


@app.route('/redfish/v1/Managers/<identity>/VirtualMedia', methods=['GET'])
@returns_json
def virtual_media_collection_resource(identity):
    if flask.request.method == 'GET':

        with Resources() as resources:

            app.logger.debug('Serving virtual media resources for '
                             'manager "%s"', identity)

            return flask.render_template(
                'virtual_media_collection.json',
                identity=identity,
                uuid=resources.managers.uuid(identity),
                devices=resources.vmedia.devices
            )


@app.route('/redfish/v1/Managers/<identity>/VirtualMedia/<device>',
           methods=['GET'])
@returns_json
def virtual_media_resource(identity, device):
    if flask.request.method == 'GET':

        with Resources() as resources:

            try:
                device_name = resources.vmedia.get_device_name(
                    identity, device)

                media_types = resources.vmedia.get_device_media_types(
                    identity, device)

                (image_name, image_url, inserted,
                 write_protected) = resources.vmedia.get_device_image_info(
                    identity, device)

            except error.FishyError as ex:
                app.logger.warning(
                    'Virtual media %s at manager %s error: '
                    '%s', device, identity, ex)
                return 'Not found', 404

        app.logger.debug('Serving virtual media %s at '
                         'manager "%s"', device, identity)

        return flask.render_template(
            'virtual_media.json',
            identity=identity,
            device=device,
            name=device_name,
            media_types=media_types,
            image_url=image_url,
            image_name=image_name,
            inserted=inserted,
            write_protected=write_protected
        )


@app.route('/redfish/v1/Managers/<identity>/VirtualMedia/<device>'
           '/Actions/VirtualMedia.InsertMedia',
           methods=['POST'])
@returns_json
def virtual_media_insert(identity, device):
    image = flask.request.json.get('Image')
    inserted = flask.request.json.get('Inserted', True)
    write_protected = flask.request.json.get('WriteProtected', False)

    with Resources() as resources:

        image_path = resources.vmedia.insert_image(
            identity, device, image, inserted, write_protected)

        for system in resources.systems.systems:
            try:
                resources.systems.set_boot_image(
                    system, device, boot_image=image_path,
                    write_protected=write_protected)

            except error.NotSupportedError as ex:
                app.logger.warning(
                    'System %s failed to set boot image %s on device %s: '
                    '%s', system, image_path, device, ex)

    app.logger.info(
        'Virtual media placed into device %s manager %s image %s '
        'inserted %s', device, identity, image or '<empty>', inserted)

    return '', 204


@app.route('/redfish/v1/Managers/<identity>/VirtualMedia/<device>'
           '/Actions/VirtualMedia.EjectMedia',
           methods=['POST'])
@returns_json
def virtual_media_eject(identity, device):
    with Resources() as resources:
        resources.vmedia.eject_image(identity, device)

        for system in resources.systems.systems:
            try:
                resources.systems.set_boot_image(system, device)

            except error.NotSupportedError as ex:
                app.logger.warning(
                    'System %s failed to remove boot image from device %s: '
                    '%s', system, device, ex)

    app.logger.info(
        'Virtual media ejected from device %s manager %s '
        'image ', device, identity)

    return '', 204


@app.route('/redfish/v1/Systems')
@returns_json
def system_collection_resource():
    with Resources() as resources:
        systems = [system for system in resources.systems.systems
                   if not instance_denied(identity=system)]

    app.logger.debug('Serving systems list')

    return flask.render_template(
        'system_collection.json', system_count=len(systems), systems=systems)


@app.route('/redfish/v1/Systems/<identity>', methods=['GET', 'PATCH'])
@ensure_instance_access
@returns_json
def system_resource(identity):

    with Resources() as resources:

        if flask.request.method == 'GET':

            app.logger.debug('Serving resources for system "%s"', identity)

            return flask.render_template(
                'system.json',
                identity=identity,
                name=resources.systems.name(identity),
                uuid=resources.systems.uuid(identity),
                power_state=resources.systems.get_power_state(identity),
                total_memory_gb=resources.systems.get_total_memory(identity),
                total_cpus=resources.systems.get_total_cpus(identity),
                boot_source_target=resources.systems.get_boot_device(identity),
                boot_source_mode=resources.systems.get_boot_mode(identity),
                managers=resources.managers.managers[:1],
                chassis=resources.chassis.chassis[:1],
                indicator_led=resources.indicators.get_indicator_state(
                    resources.systems.uuid(identity))
            )

        elif flask.request.method == 'PATCH':
            boot = flask.request.json.get('Boot')
            indicator_led_state = flask.request.json.get('IndicatorLED')
            if not boot and not indicator_led_state:
                return ('PATCH only works for Boot and '
                        'IndicatorLED elements'), 400

            if boot:
                target = boot.get('BootSourceOverrideTarget')

                if target:
                    # NOTE(lucasagomes): In libvirt we always set the boot
                    # device frequency to "continuous" so, we are ignoring the
                    # BootSourceOverrideEnabled element here

                    resources.systems.set_boot_device(identity, target)

                    app.logger.info('Set boot device to "%s" for system "%s"',
                                    target, identity)

                mode = boot.get('BootSourceOverrideMode')

                if mode:
                    resources.systems.set_boot_mode(identity, mode)

                    app.logger.info('Set boot mode to "%s" for system "%s"',
                                    mode, identity)

                if not target and not mode:
                    return ('Missing the BootSourceOverrideTarget and/or '
                            'BootSourceOverrideMode element', 400)

            if indicator_led_state:
                resources.indicators.set_indicator_state(
                    resources.systems.uuid(identity), indicator_led_state)

                app.logger.info('Set indicator LED to "%s" for system "%s"',
                                indicator_led_state, identity)

            return '', 204


@app.route('/redfish/v1/Systems/<identity>/EthernetInterfaces',
           methods=['GET'])
@ensure_instance_access
@returns_json
def ethernet_interfaces_collection(identity):
    with Resources() as resources:
        nics = resources.systems.get_nics(identity)

        return flask.render_template(
            'ethernet_interfaces_collection.json', identity=identity,
            nics=nics)


@app.route('/redfish/v1/Systems/<identity>/EthernetInterfaces/<nic_id>',
           methods=['GET'])
@ensure_instance_access
@returns_json
def ethernet_interface(identity, nic_id):
    with Resources() as resources:
        nics = resources.systems.get_nics(identity)

    for nic in nics:
        if nic['id'] == nic_id:
            return flask.render_template(
                'ethernet_interface.json', identity=identity, nic=nic)

    return 'Not found', 404


@app.route('/redfish/v1/Systems/<identity>/Actions/ComputerSystem.Reset',
           methods=['POST'])
@ensure_instance_access
@returns_json
def system_reset_action(identity):
    reset_type = flask.request.json.get('ResetType')

    with Resources() as resources:
        resources.systems.set_power_state(identity, reset_type)

    app.logger.info('System "%s" power state set to "%s"',
                    identity, reset_type)

    return '', 204


@app.route('/redfish/v1/Systems/<identity>/BIOS', methods=['GET'])
@ensure_instance_access
@returns_json
def bios(identity):
    with Resources() as resources:
        bios = resources.systems.get_bios(identity)

    app.logger.debug('Serving BIOS for system "%s"', identity)

    return flask.render_template(
        'bios.json',
        identity=identity,
        bios_current_attributes=json.dumps(bios, sort_keys=True, indent=6))


@app.route('/redfish/v1/Systems/<identity>/BIOS/Settings',
           methods=['GET', 'PATCH'])
@ensure_instance_access
@returns_json
def bios_settings(identity):

    if flask.request.method == 'GET':
        with Resources() as resources:
            bios = resources.systems.get_bios(identity)

        app.logger.debug('Serving BIOS Settings for system "%s"', identity)

        return flask.render_template(
            'bios_settings.json',
            identity=identity,
            bios_pending_attributes=json.dumps(bios, sort_keys=True, indent=6))

    elif flask.request.method == 'PATCH':
        attributes = flask.request.json.get('Attributes')

        with Resources() as resources:
            resources.systems.set_bios(identity, attributes)

        app.logger.info('System "%s" BIOS attributes "%s" updated',
                        identity, attributes)
        return '', 204


@app.route('/redfish/v1/Systems/<identity>/BIOS/Actions/Bios.ResetBios',
           methods=['POST'])
@ensure_instance_access
@returns_json
def system_reset_bios(identity):
    with Resources() as resources:
        resources.systems.reset_bios(identity)

    app.logger.info('BIOS for system "%s" reset', identity)

    return '', 204


@app.route('/redfish/v1/Systems/<identity>/SimpleStorage',
           methods=['GET'])
@ensure_instance_access
@returns_json
def simple_storage_collection(identity):
    with Resources() as resources:
        simple_storage_controllers = (
            resources.systems.get_simple_storage_collection(identity))

        return flask.render_template(
            'simple_storage_collection.json', identity=identity,
            simple_storage_controllers=simple_storage_controllers)


@app.route('/redfish/v1/Systems/<identity>/SimpleStorage/<simple_storage_id>',
           methods=['GET'])
@ensure_instance_access
@returns_json
def simple_storage(identity, simple_storage_id):
    with Resources() as resources:
        simple_storage_controllers = (
            resources.systems.get_simple_storage_collection(identity))
        try:
            storage_controller = simple_storage_controllers[simple_storage_id]
        except KeyError:
            app.logger.debug('"%s" Simple Storage resource was not found')
            return 'Not found', 404
        return flask.render_template('simple_storage.json', identity=identity,
                                     simple_storage=storage_controller)


@app.route('/redfish/v1/Systems/<identity>/Storage',
           methods=['GET'])
@ensure_instance_access
@returns_json
def storage_collection(identity):
    with Resources() as resources:
        uuid = resources.systems.uuid(identity)

        storage_col = resources.storage.get_storage_col(uuid)

        return flask.render_template(
            'storage_collection.json', identity=identity,
            storage_col=storage_col)


@app.route('/redfish/v1/Systems/<identity>/Storage/<storage_id>',
           methods=['GET'])
@ensure_instance_access
@returns_json
def storage(identity, storage_id):
    with Resources() as resources:
        uuid = resources.systems.uuid(identity)
        storage_col = resources.storage.get_storage_col(uuid)

    for stg in storage_col:
        if stg['Id'] == storage_id:
            return flask.render_template(
                'storage.json', identity=identity, storage=stg)

    return 'Not found', 404


@app.route('/redfish/v1/Systems/<identity>/Storage/<stg_id>/Drives/<drv_id>',
           methods=['GET'])
@ensure_instance_access
@returns_json
def drive_resource(identity, stg_id, drv_id):
    with Resources() as resources:
        uuid = resources.systems.uuid(identity)
        drives = resources.drives.get_drives(uuid, stg_id)

    for drv in drives:
        if drv['Id'] == drv_id:
            return flask.render_template(
                'drive.json', identity=identity, storage_id=stg_id, drive=drv)

    return 'Not found', 404


@app.route('/redfish/v1/Systems/<identity>/Storage/<storage_id>/Volumes',
           methods=['GET', 'POST'])
@ensure_instance_access
@returns_json
def volumes_collection(identity, storage_id):
    with Resources() as resources:

        uuid = resources.systems.uuid(identity)

        if flask.request.method == 'GET':

            vol_col = resources.volumes.get_volumes_col(uuid, storage_id)

            vol_ids = []
            for vol in vol_col:
                vol_id = resources.systems.find_or_create_storage_volume(vol)
                if not vol_id:
                    resources.volumes.delete_volume(uuid, storage_id, vol)
                else:
                    vol_ids.append(vol_id)

            return flask.render_template(
                'volume_collection.json', identity=identity,
                storage_id=storage_id, volume_col=vol_ids)

        elif flask.request.method == 'POST':
            data = {
                "Name": flask.request.json.get('Name'),
                "VolumeType": flask.request.json.get('VolumeType'),
                "CapacityBytes": flask.request.json.get('CapacityBytes'),
                "Id": str(os.getpid()) + datetime.now().strftime("%H%M%S")
            }
            data['libvirtVolName'] = data['Id']
            new_id = resources.systems.find_or_create_storage_volume(data)
            if new_id:
                resources.volumes.add_volume(uuid, storage_id, data)
                app.logger.debug('New storage volume created with ID "%s"',
                                 new_id)
                vol_url = ("/redfish/v1/Systems/%s/Storage/%s/"
                           "Volumes/%s" % (identity, storage_id, new_id))
                return flask.Response(status=201,
                                      headers={'Location': vol_url})


@app.route('/redfish/v1/Systems/<identity>/Storage/<stg_id>/Volumes/<vol_id>',
           methods=['GET'])
@ensure_instance_access
@returns_json
def volume(identity, stg_id, vol_id):
    with Resources() as resources:
        uuid = resources.systems.uuid(identity)
        vol_col = resources.volumes.get_volumes_col(uuid, stg_id)

        for vol in vol_col:
            if vol['Id'] == vol_id:
                vol_id = resources.systems.find_or_create_storage_volume(vol)
                if not vol_id:
                    resources.volumes.delete_volume(uuid, stg_id, vol)
                else:
                    return flask.render_template(
                        'volume.json', identity=identity, storage_id=stg_id,
                        volume=vol)

    return 'Not Found', 404


def parse_args():
    parser = argparse.ArgumentParser('sushy-emulator')
    parser.add_argument('--config',
                        type=str,
                        help='Config file path. Can also be set via '
                             'environment variable SUSHY_EMULATOR_CONFIG.')
    parser.add_argument('-i', '--interface',
                        type=str,
                        help='IP address of the local interface to listen '
                             'at. Can also be set via config variable '
                             'SUSHY_EMULATOR_LISTEN_IP. Default is all '
                             'local interfaces.')
    parser.add_argument('-p', '--port',
                        type=int,
                        help='TCP port to bind the server to.  Can also be '
                             'set via config variable '
                             'SUSHY_EMULATOR_LISTEN_PORT. Default is 8000.')
    parser.add_argument('--ssl-certificate',
                        type=str,
                        help='SSL certificate to use for HTTPS. Can also be '
                        'set via config variable SUSHY_EMULATOR_SSL_CERT.')
    parser.add_argument('--ssl-key',
                        type=str,
                        help='SSL key to use for HTTPS. Can also be set'
                        'via config variable SUSHY_EMULATOR_SSL_KEY.')
    backend_group = parser.add_mutually_exclusive_group()
    backend_group.add_argument('--os-cloud',
                               type=str,
                               help='OpenStack cloud name. Can also be set '
                                    'via environment variable OS_CLOUD or '
                                    'config variable SUSHY_EMULATOR_OS_CLOUD.'
                               )
    backend_group.add_argument('--libvirt-uri',
                               type=str,
                               help='The libvirt URI. Can also be set via '
                                    'environment variable '
                                    'SUSHY_EMULATOR_LIBVIRT_URI. '
                                    'Default is qemu:///system')

    return parser.parse_args()


def main():

    args = parse_args()

    if args.config:
        app.config.from_pyfile(args.config)

    if args.os_cloud:
        app.config['SUSHY_EMULATOR_OS_CLOUD'] = args.os_cloud

    if args.libvirt_uri:
        app.config['SUSHY_EMULATOR_LIBVIRT_URI'] = args.libvirt_uri

    else:
        for envvar in ('SUSHY_EMULATOR_LIBVIRT_URL',  # backward compatibility
                       'SUSHY_EMULATOR_LIBVIRT_URI'):
            envvar = os.environ.get(envvar)
            if envvar:
                app.config['SUSHY_EMULATOR_LIBVIRT_URI'] = envvar

    if args.interface:
        app.config['SUSHY_EMULATOR_LISTEN_IP'] = args.interface

    if args.port:
        app.config['SUSHY_EMULATOR_LISTEN_PORT'] = args.port

    if args.ssl_certificate:
        app.config['SUSHY_EMULATOR_SSL_CERT'] = args.ssl_certificate

    if args.ssl_key:
        app.config['SUSHY_EMULATOR_SSL_KEY'] = args.ssl_key

    ssl_context = None
    ssl_certificate = app.config.get('SUSHY_EMULATOR_SSL_CERT')
    ssl_key = app.config.get('SUSHY_EMULATOR_SSL_KEY')

    if ssl_certificate and ssl_key:
        ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
        ssl_context.load_cert_chain(ssl_certificate, ssl_key)

    app.run(host=app.config.get('SUSHY_EMULATOR_LISTEN_IP'),
            port=app.config.get('SUSHY_EMULATOR_LISTEN_PORT', 8000),
            ssl_context=ssl_context)

    return 0


if __name__ == '__main__':
    sys.exit(main())
